@@ -7,26 +7,39 @@ use Carp;
 use Fcntl;
 use IO::Handle;
 use IO::Socket::INET;
+use IO::Socket::UNIX;
+use List::MoreUtils qw(uniq);
 use POSIX qw(:sys_wait_h);
 use Proc::Wait3;
 use Scope::Guard;
 
 use Exporter qw(import);
 
-our $VERSION = '0.08';
-our @EXPORT_OK = qw(start_server server_ports);
+our $VERSION = '0.10';
+our @EXPORT_OK = qw(start_server restart_server server_ports);
 
 my @signals_received;
 
 sub start_server {
-    my $opts = @_ == 1 ? shift : { @_ };
-    $opts->{interval} ||= 1;
+    my $opts = {
+        (@_ == 1 ? @$_[0] : @_),
+    };
+    $opts->{interval} = 1
+        if not defined $opts->{interval};
+    $opts->{signal_on_hup} ||= 'TERM';
+    # normalize to the one that can be passed to kill
+    $opts->{signal_on_hup} =~ tr/a-z/A-Z/;
+    $opts->{signal_on_hup} =~ s/^SIG//i;
     
     # prepare args
-    my $ports = $opts->{port}
-        or croak "mandatory option ``port'' is missing\n";
+    my $ports = $opts->{port};
+    my $paths = $opts->{path};
+    croak "either of ``port'' or ``path'' option is mandatory\n"
+        unless $ports || $paths;
     $ports = [ $ports ]
-        unless ref $ports eq 'ARRAY';
+        if ! ref $ports && defined $ports;
+    $paths = [ $paths ]
+        if ! ref $paths && defined $paths;
     croak "mandatory option ``exec'' is missing or is not an arrayref\n"
         unless $opts->{exec} && ref $opts->{exec} eq 'ARRAY';
     
@@ -57,6 +70,13 @@ sub start_server {
         close $fh;
     }
     
+    # create guard that removes the status file
+    my $status_file_guard = $opts->{status_file} && Scope::Guard->new(
+        sub {
+            unlink $opts->{status_file};
+        },
+    );
+    
     print STDERR "start_server (pid:$$) starting now...\n";
     
     # start listening, setup envvar
@@ -89,6 +109,16 @@ sub start_server {
         push @sockenv, "$port=" . $sock->fileno;
         push @sock, $sock;
     }
+    for my $path (@$paths) {
+        my $sock = IO::Socket::UNIX->new(
+            Listen => Socket::SOMAXCONN(),
+            Local  => $path,
+        ) or die "failed to listen to file $path:$!";
+        fcntl($sock, F_SETFD, my $flags = '')
+            or die "fcntl(F_SETFD, 0) failed:$!";
+        push @sockenv, "$path=" . $sock->fileno;
+        push @sock, $sock;
+    }
     $ENV{SERVER_STARTER_PORT} = join ";", @sockenv;
     $ENV{SERVER_STARTER_GENERATION} = 0;
     
@@ -98,9 +128,30 @@ sub start_server {
     } for (qw/INT TERM HUP/);
     $SIG{PIPE} = 'IGNORE';
     
+    # setup status monitor
+    my ($current_worker, %old_workers);
+    my $update_status = $opts->{status_file}
+        ? sub {
+            my $tmpfn = "$opts->{status_file}.$$";
+            open my $tmpfh, '>', $tmpfn
+                or die "failed to create temporary file:$tmpfn:$!";
+            my %gen_pid = (
+                ($current_worker
+                 ? ($ENV{SERVER_STARTER_GENERATION} => $current_worker)
+                 : ()),
+                map { $old_workers{$_} => $_ } keys %old_workers,
+            );
+            print $tmpfh "$_:$gen_pid{$_}\n"
+                for sort keys %gen_pid;
+            close $tmpfh;
+            rename $tmpfn, $opts->{status_file}
+                or die "failed to rename $tmpfn to $opts->{status_file}:$!";
+        } : sub {
+        };
+    
     # the main loop
-    my $current_worker = _start_worker($opts);
-    my %old_workers;
+    $current_worker = _start_worker($opts);
+    $update_status->();
     while (1) {
         my @r = wait3(! scalar @signals_received);
         if (@r) {
@@ -111,20 +162,22 @@ sub start_server {
             } else {
                 print STDERR "old worker $died_worker died, status:$status\n";
                 delete $old_workers{$died_worker};
+                $update_status->();
             }
         }
         for (; @signals_received; shift @signals_received) {
             if ($signals_received[0] eq 'HUP') {
                 print STDERR "received HUP, spawning a new worker\n";
-                $old_workers{$current_worker} = 1;
+                $old_workers{$current_worker} = $ENV{SERVER_STARTER_GENERATION};
                 $current_worker = _start_worker($opts);
-                print STDERR "new worker is now running, sending TERM to old workers:";
+                $update_status->();
+                print STDERR "new worker is now running, sending $opts->{signal_on_hup} to old workers:";
                 if (%old_workers) {
                     print STDERR join(',', sort keys %old_workers), "\n";
                 } else {
                     print STDERR "none\n";
                 }
-                kill 'TERM', $_
+                kill $opts->{signal_on_hup}, $_
                     for sort keys %old_workers;
             } else {
                 goto CLEANUP;
@@ -134,7 +187,7 @@ sub start_server {
     
  CLEANUP:
     # cleanup
-    $old_workers{$current_worker} = 1;
+    $old_workers{$current_worker} = $ENV{SERVER_STARTER_GENERATION};
     undef $current_worker;
     print STDERR "received $signals_received[0], sending TERM to all workers:",
         join(',', sort keys %old_workers), "\n";
@@ -145,12 +198,55 @@ sub start_server {
             my ($died_worker, $status) = @r;
             print STDERR "worker $died_worker died, status:$status\n";
             delete $old_workers{$died_worker};
+            $update_status->();
         }
     }
     
     print STDERR "exitting\n";
 }
 
+sub restart_server {
+    my $opts = {
+        (@_ == 1 ? @$_[0] : @_),
+    };
+    die "--restart option requires --pid-file and --status-file to be set as well\n"
+        unless $opts->{pid_file} && $opts->{status_file};
+    
+    # get pid
+    my $pid = do {
+        open my $fh, '<', $opts->{pid_file}
+            or die "failed to open file:$opts->{pid_file}:$!";
+        my $line = <$fh>;
+        chomp $line;
+        $line;
+    };
+    
+    # function that returns a list of active generations in sorted order
+    my $get_generations = sub {
+        open my $fh, '<', $opts->{status_file}
+            or die "failed to open file:$opts->{status_file}:$!";
+        uniq sort { $a <=> $b } map { /^(\d+):/ ? ($1) : () } <$fh>;
+    };
+    
+    # wait for this generation
+    my $wait_for = do {
+        my @gens = $get_generations->()
+            or die "no active process found in the status file";
+        pop(@gens) + 1;
+    };
+    
+    # send HUP
+    kill 'HUP', $pid
+        or die "failed to send SIGHUP to the server process:$!";
+    
+    # wait for the generation
+    while (1) {
+        my @gens = $get_generations->();
+        last if scalar(@gens) == 1 && $gens[0] == $wait_for;
+        sleep 1;
+    }
+}
+
 sub server_ports {
     die "no environment variable SERVER_STARTER_PORT. Did you start the process using server_starter?",
         unless $ENV{SERVER_STARTER_PORT};
@@ -214,9 +310,7 @@ Server::Starter - a superdaemon for hot-deploying server programs
 
 =head1 DESCRIPTION
 
-It is often a pain to write a server program that supports graceful restarts, with no resource leaks.  L<Server::Starter>, solves the problem by splitting the task into two.  One is L<start_server>, a script provided as a part of the module, which works as a superdaemon that binds to zero or more TCP ports, and repeatedly spawns the server program that actually handles the necessary tasks (for example, responding to incoming commenctions).  The spawned server programs under L<Server::Starter> call accept(2) and handle the requests.
-
-The module can also be used to hot-deploy servers listening to unix domain sockets by omitting the --port option of L<start_server>.  In such case, the superdaemon will not bind to any TCP ports but instead concentrate on spawning the server program.
+It is often a pain to write a server program that supports graceful restarts, with no resource leaks.  L<Server::Starter>, solves the problem by splitting the task into two.  One is L<start_server>, a script provided as a part of the module, which works as a superdaemon that binds to zero or more TCP ports or unix sockets, and repeatedly spawns the server program that actually handles the necessary tasks (for example, responding to incoming commenctions).  The spawned server programs under L<Server::Starter> call accept(2) and handle the requests.
 
 To gracefully restart the server program, send SIGHUP to the superdaemon.  The superdaemon spawns a new server program, and if (and only if) it starts up successfully, sends SIGTERM to the old server program.
 
@@ -233,7 +327,7 @@ A Net::Server personality that can be run under L<Server::Starter> exists under
 
 =item server_ports
 
-Returns zero or more file descriptors on which the server program should call accept(2) in a hashref.  Each element of the hashref is: (host:port|port)=file_descriptor.
+Returns zero or more file descriptors on which the server program should call accept(2) in a hashref.  Each element of the hashref is: (host:port|port|path_of_unix_socket) => file_descriptor.
 
 =item start_server
 
@@ -243,8 +337,7 @@ Starts the superdaemon.  Used by the C<start_server> scirpt.
 
 =head1 AUTHOR
 
-Kazuho Oku E<lt>kazuhooku@gmail.comE<gt>
-Copyright (C) 2009-2010 Cybozu Labs, Inc.
+Kazuho Oku
 
 =head1 SEE ALSO
 
